/****************************************************************************/
// camshow.c
//
//  - functions to fetch, convert, display, & rotate thumbnails
//  - sidebar window initialization & maintenance
//
/****************************************************************************/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Cameraderie.
 *
 * The Initial Developer of the Original Code is Richard L. Walsh.
 * 
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Heiko Nitzsche (gbm access code copied from gbmtest.c)
 *
 * ***** END LICENSE BLOCK ***** */

/****************************************************************************/

#include "camapp.h"
#include "camusb.h"
#include <stddef.h>
#include <fcntl.h>
#include <io.h>
#include "gbm.h"

/****************************************************************************/

#define CAM_MAGIC       0x6b7c8d9e
#define CAM_MAGICSTR    "0x6b7c8d9e"
#define CAM_MAGICJPG    "0x6b7c8d9e.JPG"

typedef struct _CAMFD {
    char *          pBuf;
    LONG            lSize;
    LONG            lPos;
} CAMFD;

/* bitmap header incl. color table */
/* for 8 bit and lower images */
#pragma pack(2)
typedef struct
{
   BITMAPINFOHEADER2 bmp2;
   RGB2 argb2Color[0x100];
} BMP8HEADER;

/* for above 8 bit images */
typedef struct
{
   BITMAPINFOHEADER2 bmp2;
} BMP24HEADER;
#pragma pack()

/****************************************************************************/
/*
void        InitShow( HWND hwnd);
void        TermShow( HWND hwnd);
void        ShowThumbnail( HWND hwnd, CAMRecPtr pcr);
void        SetThumbnailTitle( HWND hwnd, char * pszTitle);
void        UpdateThumbnail( HWND hwnd);
void        RotateThumbnail( HWND hwnd, ULONG ulCmd);
*/

MRESULT _System SidebarSubProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);

static HBITMAP  cam_create_bmp( const char *filename, ULONG ulRotate);
static HBITMAP  cam_rotate_bmp( HPS hps, HBITMAP hBmpIn,
                                PBITMAPINFOHEADER2 bmpInfo, ULONG ulRotate);

static int  GBMENTRY cam_open(const char *fn, int mode);
static int  GBMENTRY cam_create(const char *fn, int mode);
static void GBMENTRY cam_close(int fd);
static long GBMENTRY cam_lseek( int fd, long pos, int whence);
static int  GBMENTRY cam_read(int fd, void *buf, int len);
static int  GBMENTRY cam_write(int fd, const void *buf, int len);
static int  get_checked_internal_open_mode(int mode);
static int  get_checked_internal_lseek_mode(const int whence);
static int  isReadSupportedFormat(const int ft, const int bpp);

/****************************************************************************/

static BOOL     fSlowMachine = FALSE;
static HBITMAP  hbmpBlank = 0;
static PFNWP    pfnSidebar = 0;
static CAMFD    camFD = {0,0,0};

/*
static MATRIXLF mRot0     = { 0x00010000, 0,          0,     //   1   0  0
                              0,          0x00010000, 0,     //   0   1  0
                              0,          0,          1 };   //   0   0  1
*/
static MATRIXLF mRot90    = { 0,          0x00010000, 0,     //   0   1  0
                              0xffff0000, 0,          0,     //  -1   0  0
                              120,        0,          1 };   // 120   0  1

static MATRIXLF mRot180   = { 0xffff0000, 0,          0,     //  -1   0  0
                              0,          0xffff0000, 0,     //   0  -1  0
                              160,        120,        1 };   // 160 120  1

static MATRIXLF mRot270   = { 0,          0xffff0000, 0,     //   0  -1  0
                              0x00010000, 0,          0,     //   1   0  0
                              0,          160,        1 };   //   0 160  1

static CAMNLS   nlsSidebar[] = {
    { IDC_EDITGRP,      BTN_EditGroups },
    { IDC_GRPFIRST,     GRP_Default },
    { 0,                0 }};


/****************************************************************************/

void        InitShow( HWND hwnd)

{
    HPS         hps;
    HWND        hSidebar;
    HWND        hTemp;
    SWP         swpInfo;

    InitGroups( hwnd);

    hSidebar = WinWindowFromID( hwnd, IDC_SIDEBAR);
    LoadDlgStrings( hSidebar, nlsSidebar);

    // the thumbnail window seems to ignore SM_SETHANDLE and won't
    // display new bitmaps if it is created with the bitmap style,
    // so we have to change the style & load the bitmap manually
    hps = WinGetPS( hSidebar);
    if (hps)
    {
        hbmpBlank = GpiLoadBitmap( hps, 0, IDB_THUMBBMP, 0, 0);
        WinReleasePS( hps);
    }

    hTemp = WinWindowFromID( hSidebar, IDC_BMP);
    WinSetWindowBits( hTemp, QWL_STYLE, SS_BITMAP, SS_TEXT | SS_BITMAP);
    WinSendMsg( hTemp, SM_SETHANDLE, (MP)hbmpBlank, 0);

    WinQueryWindowPos( hTemp, &swpInfo);

    hTemp = WinWindowFromID( hSidebar, IDC_ROTCC);
    swpInfo.y += 6;
    WinSetWindowPos( hTemp, 0, 8, swpInfo.y,
                     16, 16, SWP_MOVE | SWP_SIZE);

    hTemp = WinWindowFromID( hSidebar, IDC_ROTCW);
    swpInfo.x += swpInfo.cx - 16 - 8;
    WinSetWindowPos( hTemp, 0, swpInfo.x, swpInfo.y,
                     16, 16, SWP_MOVE | SWP_SIZE);

    pfnSidebar = WinSubclassWindow( hSidebar, SidebarSubProc);

    gbm_init();
    gbm_io_setup( cam_open,
                  cam_create,
                  cam_close,
                  cam_lseek,
                  cam_read,
                  cam_write);

    return;
}

/****************************************************************************/

MRESULT _System SidebarSubProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
 
{

    if (msg == WM_WINDOWPOSCHANGED &&
        (((PSWP)mp1)->fl & SWP_SIZE) &&
        ((PSWP)mp1)[0].cy != ((PSWP)mp1)[1].cy) {

        HWND    hTemp;
        LONG    lChg;
        SWP     aswp[5];

        lChg = ((PSWP)mp1)[0].cy - ((PSWP)mp1)[1].cy;

        hTemp = WinWindowFromID( hwnd, IDC_BMP);
        WinQueryWindowPos( hTemp, &aswp[0]);
        aswp[0].y += lChg;
        aswp[0].fl = SWP_MOVE;

        hTemp = WinWindowFromID( hwnd, IDC_INFO);
        WinQueryWindowPos( hTemp, &aswp[1]);
        aswp[1].y += lChg;
        aswp[1].fl = SWP_MOVE;

        hTemp = WinWindowFromID( hwnd, IDC_ROTCW);
        WinQueryWindowPos( hTemp, &aswp[2]);
        aswp[2].y += lChg;
        aswp[2].fl = SWP_MOVE;

        hTemp = WinWindowFromID( hwnd, IDC_ROTCC);
        WinQueryWindowPos( hTemp, &aswp[3]);
        aswp[3].y += lChg;
        aswp[3].fl = SWP_MOVE;

        hTemp = WinWindowFromID( hwnd, IDC_EDITGRP);
        WinQueryWindowPos( hTemp, &aswp[4]);
        aswp[4].y += lChg;
        aswp[4].fl = SWP_MOVE | ((aswp[2].y < 3) ? SWP_HIDE : SWP_SHOW);

        WinSetMultWindowPos( 0, aswp, 5);
        RepositionGroupButtons( hwnd, lChg);
    }

    return (pfnSidebar( hwnd, msg, mp1, mp2));
}

/****************************************************************************/

void        TermShow( HWND hwnd)

{
    HBITMAP     hBmp;
    HWND        hThumbWnd;
    HWND        hSidebar;

    hSidebar = WinWindowFromID( hwnd, IDC_SIDEBAR);

    if (camFD.pBuf) {
        free( camFD.pBuf);
        camFD.pBuf = 0;
    }

    gbm_deinit();

    hThumbWnd = WinWindowFromID( hSidebar, IDC_BMP);
    hBmp = (HBITMAP)WinSendMsg( hThumbWnd, SM_QUERYHANDLE, 0, 0);
    WinSendMsg( hThumbWnd, SM_SETHANDLE, 0, 0);

    if (hBmp && hBmp != hbmpBlank)
        GpiDeleteBitmap( hBmp);

    if (hbmpBlank) {
        GpiDeleteBitmap( hbmpBlank);
        hbmpBlank = 0;
    }

    return;
}

/****************************************************************************/

// ShowThumbnail() is designed to deal with the possibility that it may
// be called recursively while waiting for DispatchUsbMsg() to return.
// This could occur if DispatchUsbMsg()'s msg loop dispatched a msg which
// caused the container's cursor to change.
//
// pCur contains the record currently being processed.  If it is non-zero
// when the function is invoked, then the call is recursive.  The new
// request is stored in pNew, then the function returns.  Subsequent
// recursive calls overwrite pNew so it always contains the most recent
// request.  When DispatchUsbMsg() returns, the function finishes its
// handling of the current thumbnail, then loops to start servicing the
// most recent request in pNew.  This ensures that when the stream of
// requests stops, the display will be in sync with the currently
// selected record.

void    ShowThumbnail( HWND hwnd, CAMRecPtr pcr)

{
static CAMRecPtr pCur = 0;
static CAMRecPtr pNew = 0;
    HBITMAP     hbmpOld;
    HBITMAP     hbmpNew;
    HWND        hSidebar;
    HWND        hStatic;
    char *      ptr;
    QMSG        qmsg;
    char        szText[32];

    // save the request in case this call is recursive
    if (pNew)
        printf( "handle %d skipped\n", (int)pCur->hndl);
    pNew = pcr;

    // this discards all but the newest request
    while (WinPeekMsg( 0, &qmsg, hwnd, CAMMSG_SHOWTHUMB, CAMMSG_SHOWTHUMB,
                       PM_REMOVE))
        pNew = (CAMRecPtr)qmsg.mp1;

    // this is a recursive call, so exit
    if (pCur)
        return;

    // do things that only need to be done once
    hSidebar = WinWindowFromID( hwnd, IDC_SIDEBAR);
    hStatic = WinWindowFromID( hSidebar, IDC_BMP);

    // this buffer should have been freed long ago
    if (camFD.pBuf) {
        printf( "ShowThumbnail - found pBuf on entry\n");
        free( camFD.pBuf);
        camFD.pBuf = 0;
    }

    while (pNew) {

        pCur = pNew;
        pNew = 0;
        hbmpNew = 0;

        // no thumbnail available
        if (pCur->tnsize == 0) {
            hbmpNew = hbmpBlank;
            UpdateStatus( hwnd, STS_NoThumbnail);
        }
        else {
            // get the thumbnail from the camera 
            if (DispatchUsbMsg( CAMMSG_GETTHUMB, pCur->hndl,
                                (void**)&camFD.pBuf))
                camcli_error( "Could not get thumbnail for handle %d",
                              (int)pCur->hndl);

            // if another request is pending and the user chose this
            // option, discard the data & start on the next request
            if (pNew && fSlowMachine) {
                printf( "handle %d discarded\n", (int)pCur->hndl);
                if (camFD.pBuf) {
                    free( camFD.pBuf);
                    camFD.pBuf = 0;
                }
                continue;
            }

            // turn the data into a bitmap
            if (camFD.pBuf) {

                // setup this unfortunate method of passing data to gbm
                camFD.lSize = (LONG)pCur->tnsize;
                camFD.lPos  = 0;

                // our gbm_io_open() replacement recognizes a camera
                // request by this (hopefully unique) filename
                if (pCur->tnfmtnbr == PTP_OFC_JFIF ||
                    pCur->tnfmtnbr == PTP_OFC_EXIF_JPEG)
                    strcpy( szText, CAM_MAGICJPG);
                else {
                    strcpy( szText, CAM_MAGICSTR);
                    ptr = strrchr( pCur->title, '.');
                    if (ptr)
                        strcat( szText, ptr);
                }

                // create the bitmap and rotate it if needed
                hbmpNew = cam_create_bmp( szText, STAT_GETROT(pCur->status));
            }

            // if either FetchThumbnail() or cam_create_bmp() failed,
            // revert to the blank bitmap
            if (!hbmpNew) {
                hbmpNew = hbmpBlank;
                UpdateStatus( hwnd, STS_GetThumbErr);
            }

            // our gbm_io_close() replacement should already have freed
            // the buffer returned by GetThumb(), but check to make sure
            if (camFD.pBuf) {
                free( camFD.pBuf);
                camFD.pBuf = 0;
            }
        }

        // display the new bitmap, destroy the old one, and attach
        // pCur to the thumbnail window for use by RotateThumbnail()
        if (hbmpNew) {
            WinSetWindowULong( hStatic, QWL_USER, (ULONG)pCur);
            SetThumbnailTitle( hwnd, pCur->title);
            hbmpOld = (HBITMAP)WinSendMsg( hStatic, SM_QUERYHANDLE, 0, 0);
            WinSendMsg( hStatic, SM_SETHANDLE, (MP)hbmpNew, 0);
            if (hbmpOld && hbmpOld != hbmpBlank)
                GpiDeleteBitmap( hbmpOld);
        }

    } // while pNew

    // there are no requests in process or pending
    pCur = 0;
    return;
}

/****************************************************************************/

void    SetThumbnailTitle( HWND hwnd, char * pszTitle)

{
    char *  ptr;
    char    szText[CCHMAXPATH];

    ptr = strrchr( pszTitle, '\\');
    if (ptr)
        ptr++;
    else {
        ptr = pszTitle;
        if (ptr[1] == ':')
            ptr += 2;
    }

    strcpy( szText, ptr);

    ptr = szText;
    while ((ptr = strpbrk( ptr, "\r\n")) != 0)
        strcpy( ptr, ptr + strspn( ptr, "\r\n"));

    WinSetDlgItemText( WinWindowFromID( hwnd, IDC_SIDEBAR), IDC_INFO, szText);

    return;
}

/****************************************************************************/

// after deleting files, the current thumbnail may be invalid because the
// selection has changed or there aren't any records left;  this displays
// the t/n for the current record or the blank bitmap as appropriate

void    UpdateThumbnail( HWND hwnd)

{
    HBITMAP         hbmpOld;
    HWND        	hSidebar;
    HWND        	hStatic;
    CAMRecPtr       pcr;
    CAMRecPtr       pcrOld;

do {
    hSidebar = WinWindowFromID( hwnd, IDC_SIDEBAR);
    hStatic = WinWindowFromID( hSidebar, IDC_BMP);

    pcrOld = (CAMRecPtr)WinQueryWindowULong( hStatic, QWL_USER);
    pcr = (CAMRecPtr)WinSendDlgItemMsg( hwnd, FID_CLIENT,
                    CM_QUERYRECORDEMPHASIS, (MP)CMA_FIRST, (MP)CRA_CURSORED);

    if (pcr && pcr != (CAMRecPtr)-1) {
        if (pcr != pcrOld) {
            WinSendDlgItemMsg( hwnd, FID_CLIENT, CM_SETRECORDEMPHASIS,
                               (MP)pcr, MPFROM2SHORT( TRUE, CRA_SELECTED));
            WinPostMsg( hwnd, CAMMSG_SHOWTHUMB, (MP)pcr, 0);
        }
        break;
    }

    WinSetWindowULong( hStatic, QWL_USER, 0);
    WinSetDlgItemText( hSidebar, IDC_INFO, "");
    UpdateStatus( hwnd, STS_NoFiles);
    hbmpOld = (HBITMAP)WinSendMsg( hStatic, SM_QUERYHANDLE, 0, 0);
    if (hbmpOld != hbmpBlank) {
        WinSendMsg( hStatic, SM_SETHANDLE, (MP)hbmpBlank, 0);
        if (hbmpOld)
            GpiDeleteBitmap( hbmpOld);
    }

} while (0);

    return;
}

/****************************************************************************/

void    RotateThumbnail( HWND hwnd, ULONG ulCmd)

{
    HWND        hSidebar;
    HWND        hStatic;
    HBITMAP     hBmp;
    HBITMAP     hBmpRot;
    HDC         hdc = 0;
    HPS         hps = 0;
    ULONG       ulRot;
    PULONG      pulStatus;
    SIZEL       sizl;
    CAMRecPtr   pcr;
    BITMAPINFOHEADER2   bmih;

do {
    if (ulCmd == IDC_ROTCC)
        ulRot = STAT_ROT90;
    else
    if (ulCmd == IDC_ROTCW)
        ulRot = STAT_ROT270;
    else
        break;

    hSidebar = WinWindowFromID( hwnd, IDC_SIDEBAR);
    hStatic = WinWindowFromID( hSidebar, IDC_BMP);
    pcr = (CAMRecPtr)WinQueryWindowULong( hStatic, QWL_USER);
    if (!pcr)
        break;

    hBmp = (HBITMAP)WinSendMsg( hStatic, SM_QUERYHANDLE, 0, 0);
    if (hBmp == 0 || hBmp == hbmpBlank)
        break;

    memset( &bmih, 0, sizeof(bmih));
    bmih.cbFix = sizeof(bmih);

    if (!GpiQueryBitmapInfoHeader( hBmp, &bmih)) {
        printf( "GpiQueryBitmapInfoHeader failed\n");
        break;
    }

    hdc = DevOpenDC( 0, OD_MEMORY, "*", 0L, 0, 0);
    if (hdc == 0)
    {
        printf("DevOpenDC failed\n");
        break;
    }

    sizl.cx = ((bmih.cx >= bmih.cy) ? bmih.cx : bmih.cy);
    sizl.cy = sizl.cx;

    hps = GpiCreatePS( 0, hdc, &sizl, PU_PELS | GPIF_DEFAULT | GPIT_MICRO | GPIA_ASSOC);
    if (hps == 0)
    {
        printf("GpiCreatePS failed\n");
        break;
    }

    hBmpRot = cam_rotate_bmp( hps, hBmp, &bmih, ulRot);
    if (!hBmpRot)
        break;

    WinSendMsg( hStatic, SM_SETHANDLE, (MP)hBmpRot, 0);
    GpiDeleteBitmap( hBmp);

    pulStatus = &(pcr->status);
    if (ulCmd == IDC_ROTCC)
        *pulStatus = STAT_SETROTCC(*pulStatus);
    else
        *pulStatus = STAT_SETROTCW(*pulStatus);

} while (0);

    if (hps) {
        GpiSetBitmap( hps, 0);
        GpiDestroyPS( hps);
    }
    if (hdc)
        DevCloseDC( hdc);

    return;
}

/****************************************************************************/
/****************************************************************************/

static HBITMAP cam_create_bmp( const char *filename, ULONG ulRotate)
{
    int         i;
    int         ftr;
    int         dataSize;
    GBM         gbm;
    GBMRGB      gbmrgb[0x100];
    BMP8HEADER  header8BMP;
    BMP24HEADER header24BMP;
    SIZEL       sizl;
    BITMAPINFO2 *bmpInfo;

    int         fd = -1;
    byte  *     data = NULL;
    HBITMAP     hBmp = 0;
    HDC         hdc = 0;
    HPS         hps = 0;

do {
    /* check if file format is supported */
    if (gbm_guess_filetype(filename, &ftr) != GBM_ERR_OK)
    {
        printf("Input format is not supported.\n");
        break;
    }
 
    /* Read bitmap file */
 
    /* open bitmap file */
    fd = gbm_io_open(filename, GBM_O_RDONLY);
    if (fd < 0)
    {
        printf("FAILED, Cannot open file.\n");
        break;
    }
 
    /* read bitmap header */
    if (gbm_read_header(filename, fd, ftr, &gbm, "") != GBM_ERR_OK)
    {
        printf("FAILED, Cannot read file header.\n");
        break;
    }
 
    /* check whether we can support the provided color depth (restrict to algorithms) */
    if (!isReadSupportedFormat(ftr, gbm.bpp))
    {
        printf("FAILED: Input format color depth is not supported.\n");
        break;
    }
 
    /* read bitmap palette */
    memset(gbmrgb, 0, sizeof(gbmrgb));
    if (gbm_read_palette(fd, ftr, &gbm, gbmrgb) != GBM_ERR_OK)
    {
        printf("FAILED, Cannot read bitmap palette.\n");
        break;
    }
 
    /* calc needed memory and load bitmap data */
    dataSize = (((gbm.w * gbm.bpp + 31) / 32) * 4) * gbm.h;
    if ((data = (byte *) malloc(dataSize)) == NULL)
    {
        printf("FAILED, Out of memory.\n");
        break;
    }
    memset(data, 0, dataSize);
 
    /* read bitmap data */
    if (gbm_read_data(fd, ftr, &gbm, data) != GBM_ERR_OK)
    {
        printf("FAILED, Cannot read bitmap data.\n");
        break;
    }
 
    gbm_io_close(fd);
    fd = -1;
 
    /* create bitmap header */
    if (gbm.bpp <= 8)
    {
        bmpInfo = (PBITMAPINFO2)&header8BMP;
        memset( &header8BMP, 0, sizeof(BMP8HEADER));

        header8BMP.bmp2.cbFix           = sizeof(BITMAPINFOHEADER2);
        header8BMP.bmp2.cx              = gbm.w;
        header8BMP.bmp2.cy              = gbm.h;
        header8BMP.bmp2.cPlanes         = 1;                /*  Number of bit planes. */
        header8BMP.bmp2.cBitCount       = gbm.bpp;          /*  Number of bits per pel within a plane. */
        header8BMP.bmp2.ulCompression   = BCA_UNCOMP;       /*  Compression scheme used to store the bit map. */
        header8BMP.bmp2.cbImage         = dataSize;         /*  Length of bit-map storage data, in bytes. */

        /* copy color table */
        if ((gbm.bpp == 1) &&
            ((gbmrgb[1].b == 0)    && (gbmrgb[1].r == 0)    && (gbmrgb[1].g == 0)  &&
             (gbmrgb[0].b == 0xff) && (gbmrgb[0].r == 0xff) && (gbmrgb[0].g == 0xff)))
        {
            /* special handling of BW bitmaps due to the way PM works */
            header8BMP.argb2Color[1].bBlue     = 0xff;
            header8BMP.argb2Color[1].bRed      = 0xff;
            header8BMP.argb2Color[1].bGreen    = 0xff;
        }
        else
            for (i = 0; i < 256; i++)
            {
                header8BMP.argb2Color[i].bBlue     = gbmrgb[i].b;
                header8BMP.argb2Color[i].bRed      = gbmrgb[i].r;
                header8BMP.argb2Color[i].bGreen    = gbmrgb[i].g;
            }
    }
    else
    {
        bmpInfo = (PBITMAPINFO2) &header24BMP;
        memset( &header24BMP, 0, sizeof(BMP24HEADER));

        header24BMP.bmp2.cbFix           = sizeof(BITMAPINFOHEADER2);
        header24BMP.bmp2.cx              = gbm.w;
        header24BMP.bmp2.cy              = gbm.h;
        header24BMP.bmp2.cPlanes         = 1;                /*  Number of bit planes. */
        header24BMP.bmp2.cBitCount       = gbm.bpp;          /*  Number of bits per pel within a plane. */
        header24BMP.bmp2.ulCompression   = BCA_UNCOMP;       /*  Compression scheme used to store the bit map. */
        header24BMP.bmp2.cbImage         = dataSize;         /*  Length of bit-map storage data, in bytes. */
    }

   /* create system bitmap */

    /* create a memory context */
    hdc = DevOpenDC( 0, OD_MEMORY, "*", 0L, 0, 0);
    if (hdc == 0)
    {
        printf("FAILED, DevOpenDC\n");
        break;
    }

    sizl.cx = ((bmpInfo->cx >= bmpInfo->cy) ? bmpInfo->cx : bmpInfo->cy);
    sizl.cy = sizl.cx;

    /* create presentation space */
    hps = GpiCreatePS( 0, hdc, &sizl, PU_PELS | GPIF_DEFAULT | GPIT_MICRO | GPIA_ASSOC);
    if (hps == 0)
    {
        printf("FAILED, GpiCreatePS\n");
        break;
    }

    /* create the bitmap */
    hBmp = GpiCreateBitmap( hps, (PBITMAPINFOHEADER2)bmpInfo,
                            CBM_INIT, (PBYTE) data, bmpInfo );

    if (ulRotate) {
        HBITMAP hBmpRot = cam_rotate_bmp( hps, hBmp,
                                          (PBITMAPINFOHEADER2)bmpInfo,
                                          ulRotate);
        if (hBmpRot) {
            GpiDeleteBitmap( hBmp);
            hBmp = hBmpRot;
        }
    }

} while (0);

    if (hps) {
        GpiSetBitmap( hps, 0);
        GpiDestroyPS( hps);
    }
    if (hdc)
        DevCloseDC( hdc);
    if (data)
        free( data);
    if (fd >= 0)
        gbm_io_close( fd);

    return hBmp;
}

/****************************************************************************/

static HBITMAP cam_rotate_bmp( HPS hps, HBITMAP hBmpIn,
                               PBITMAPINFOHEADER2 bmpInfo, ULONG ulRotate)
{
    MATRIXLF *  pRot;
    HBITMAP     hBmpOut = 0;
    HBITMAP     hBmpWork = 0;
    POINTL      aptl[4];

do {
    memset( aptl, 0, sizeof(aptl));
    aptl[1].x = bmpInfo->cx;
    aptl[1].y = bmpInfo->cy;
    aptl[3].x = bmpInfo->cx;
    aptl[3].y = bmpInfo->cy;

    if (ulRotate == STAT_ROT0)
        break;

    if (ulRotate == STAT_ROT180) {
        pRot = &mRot180;
        mRot180.lM31 = bmpInfo->cx;
        mRot180.lM32 = bmpInfo->cy;
    }
    else {
        bmpInfo->cx = aptl[1].y;
        bmpInfo->cy = aptl[1].x;

        if (ulRotate == STAT_ROT270) {
            pRot = &mRot270;
            mRot270.lM32 = bmpInfo->cy;
        }
        else {
            pRot = &mRot90;
            mRot90.lM31 = bmpInfo->cx;
        }
    }

    GpiSetBitmap( hps, 0);

    hBmpWork = GpiCreateBitmap( hps, bmpInfo, 0, 0, 0);
    if (hBmpWork == 0) {
        printf( "GpiCreateBitmap failed\n");
        break;
    }

    if (GpiSetBitmap( hps, hBmpWork) == HBM_ERROR) {
        printf( "GpiSetBitmap failed\n");
        break;
    }

    if (!GpiSetModelTransformMatrix( hps, 9, pRot, TRANSFORM_REPLACE)) {
        printf( "Rotate failed\n");
        break;
    }

    if (GpiWCBitBlt( hps, hBmpIn, 4, aptl, ROP_SRCCOPY,
                     BBO_IGNORE) == GPI_ERROR) {
        printf( "GpiWCBitBlt failed\n");
        break;
    }

    hBmpOut = hBmpWork;

} while (0);

    if (hBmpOut == 0) {
        if (hBmpWork)
            GpiDeleteBitmap( hBmpWork);
    }

    return hBmpOut;
}

/****************************************************************************/
/****************************************************************************/

static int GBMENTRY cam_open(const char *fn, int mode)

{
    if (strtoul( fn, 0, 0) != CAM_MAGIC)
       return open( fn, get_checked_internal_open_mode( mode));

    if (camFD.pBuf == 0)
        return -1;

    camFD.lPos = 0;

    return (int)CAM_MAGIC;
}

static int GBMENTRY cam_create(const char *fn, int mode)

{
    if (strtoul( fn, 0, 0) == CAM_MAGIC)
        return -1;

    return open(fn, O_CREAT | O_TRUNC | get_checked_internal_open_mode(mode), S_IREAD | S_IWRITE);
}

static void GBMENTRY cam_close(int fd)

{ 
    if (fd != (int)CAM_MAGIC)
       close( fd);
    else {
        free( camFD.pBuf);
        camFD.pBuf = 0;
    }

    return;
}

static long GBMENTRY cam_lseek( int fd, long pos, int whence)

{
    LONG    lCur;

    if (fd != (int)CAM_MAGIC)
        return lseek( fd, pos, get_checked_internal_lseek_mode(whence));

    switch (whence) {

        case GBM_SEEK_SET:
            lCur = pos;
            break;

        case GBM_SEEK_CUR:
            lCur = camFD.lPos + pos;
            break;

        case GBM_SEEK_END:
            lCur = camFD.lSize + pos;
            break;

        default:
            lCur = -1;
    }

    if (lCur < 0 || lCur > camFD.lSize)
        lCur = -1;
    else
        camFD.lPos = lCur;

    return lCur;
}

static int GBMENTRY cam_read(int fd, void *buf, int len)

{ 
    LONG    lRead;

    if (!buf || len <= 0)
        return -1;

    if (fd != (int)CAM_MAGIC)
        return read(fd, buf, len);

    if (camFD.lPos + len > camFD.lSize)
        lRead = camFD.lSize - camFD.lPos;
    else
        lRead = len;

    memcpy( buf, &camFD.pBuf[camFD.lPos], lRead);
    camFD.lPos += lRead;

    return lRead;
}

static int GBMENTRY cam_write(int fd, const void *buf, int len)

{
    if (fd != (int)CAM_MAGIC)
        return write(fd, buf, len);

    return -1;
}

/****************************************************************************/

/* map supported GBM_O_* defines to compiler specific ones */
static int get_checked_internal_open_mode(int mode)
{
   const static int supported_open_modes = GBM_O_RDONLY |
                                           GBM_O_WRONLY |
                                           GBM_O_RDWR   |
                                           GBM_O_EXCL   |
                                           GBM_O_NOINHERIT;

   /* internal, compiler specific open mode */
   int open_mode = 0;

   /* map supported public defines to compiler specific ones */

   /* check if only supported modes are provided */
   if (!(mode & supported_open_modes))
   {
      return 0xffffffff;
   }

   /* mask external binary mode bit */
   mode &= ~GBM_O_BINARY;

   if (mode & GBM_O_RDONLY)
   {
      open_mode |= O_RDONLY;
   }
   else if (mode & GBM_O_WRONLY)
   {
      open_mode |= O_WRONLY;
   }
   else if (mode & GBM_O_RDWR)
   {
      open_mode |= O_RDWR;
   }
   else if (mode & GBM_O_EXCL)
   {
      open_mode |= O_EXCL;
   }
   else
   {
      return 0xffffffff;
   }

   if (mode & GBM_O_NOINHERIT)
   {
      open_mode |= O_NOINHERIT;
   }

   /* force binary mode if necessary */
   #ifdef O_BINARY
     open_mode |= O_BINARY;
   #endif

   return open_mode;
}

/* map supported GBM_SEEK_* defines to compiler specific ones */
static int get_checked_internal_lseek_mode( const int whence)

{
   int internal_whence = -1;

   switch(whence)
   {
      case GBM_SEEK_SET:
         internal_whence = SEEK_SET;
         break;

      case GBM_SEEK_CUR:
         internal_whence = SEEK_CUR;
         break;

      case GBM_SEEK_END:
         internal_whence = SEEK_END;
         break;

      default:
         /* as we provide an illegal whence value, the OS takes care
          * for correct error reporting
          */
         break;
   }

   return internal_whence;
}

/****************************************************************************/

static int isReadSupportedFormat( const int ft, const int bpp)

{
   GBMFT readFormat;
   int   readFormatFlag = 0;

   switch (bpp)
   {
      case 64: readFormatFlag = GBM_FT_R64; break;
      case 48: readFormatFlag = GBM_FT_R48; break;
      case 32: readFormatFlag = GBM_FT_R32; break;
      case 24: readFormatFlag = GBM_FT_R24; break;
      case  8: readFormatFlag = GBM_FT_R8;  break;
      case  4: readFormatFlag = GBM_FT_R4;  break;
      case  1: readFormatFlag = GBM_FT_R1;  break;
      default: readFormatFlag = 0;          break;
   }

   if (gbm_query_filetype( ft, &readFormat) == GBM_ERR_OK &&
       (readFormat.flags & readFormatFlag))
       return 1;

   return 0;
}

/****************************************************************************/

